package kms

import (
	"context"
	"encoding/base64"
	"fmt"
	"net"
	"time"

	pbIC "code.fbi.h-da.de/danet/proto-kms/api/gen/proto/go/kmsintercom"
	"github.com/google/uuid"
	"github.com/sirupsen/logrus"
	log "github.com/sirupsen/logrus"
	"google.golang.org/grpc"
	"google.golang.org/grpc/credentials/insecure"
)

type KmsPeerStatus int16

const (
	KmsPeerUp      KmsPeerStatus = iota // peer was seen up lately
	KmsPeerDown                         // peer not reachable anymore
	KmsPeerPending                      // peer was added and has to be succesfully contacted once
	KmsPeerUnknown                      // not known, not initialized
)

type kmsPeerInfo interface {
	GetKmsPeerStatus() KmsPeerStatus
	GetKmsPeerId() uuid.UUID
	GetKmsPeerQkdiId() uint32
	KmsPeerKeyInit()
}

type kmsPeer struct {
	externalNotifierKMSPeer chan string
	peerClient              pbIC.KmsTalkerClient
	peerStatus              KmsPeerStatus
	servingQLE              *QuantumElement
	tcpSocket               net.TCPAddr // the IP address and TCP port (aka socket) of the kms peer
	tcpSocketStr            string      // string rep. of tcpSocket
	et                      CryptoAlgorithm
	name                    string    // the name of the kms peer
	id                      uuid.UUID // uuid of the peer
	quit                    chan bool // cancel the peer goroutine
}

func NewKmsPeer(servQLE *QuantumElement, tcpSocketStr string, in chan string) (peer *kmsPeer, err error) {
	newPeerConn, err := grpc.Dial(tcpSocketStr, grpc.WithTransportCredentials(insecure.NewCredentials()))
	if err != nil {
		return nil, err
	}
	peerClient := pbIC.NewKmsTalkerClient(newPeerConn)

	return &kmsPeer{
		externalNotifierKMSPeer: in,
		peerClient:              peerClient,
		peerStatus:              KmsPeerUnknown,
		servingQLE:              servQLE,
		tcpSocketStr:            tcpSocketStr,
		et:                      NewAES(),
		id:                      uuid.New(),
		quit:                    make(chan bool),
	}, nil
}

// Handles everything with respect to a specific KMS peer
//func (ph *kmsPeer) PeerHandler(kmsName string) {
//	//	log.Infof("%s started PeerHandler for %s:", kmsName, ph.tcpSocketStr)
//
//	// contact peer
//	newPeerConn, err := grpc.Dial(ph.tcpSocketStr, grpc.WithTransportCredentials(insecure.NewCredentials()))
//	if err != nil {
//		log.Errorf("%s: did not connect: %v", kmsName, err)
//		ph.peerStatus = KmsPeerDown
//		return
//	}
//	defer newPeerConn.Close()
//
//	c := pb.NewKmsETSIClient(newPeerConn)
//
//	// Contact the server and print out its response.
//	ctx, cancel := context.WithTimeout(context.Background(), time.Second)
//	defer cancel()
//	r, err := c.ETSICapabilities(ctx, &pb.ETSICapabilitiesRequest{MyKmsName: kmsName})
//	if err != nil {
//		log.Errorf("%s: could not greet: %v", kmsName, err)
//		ph.peerStatus = KmsPeerDown
//		// Send notification about change
//		if ph.externalNotifierKMSPeer != nil {
//			ph.externalNotifierKMSPeer <- ph.tcpSocketStr
//		}
//	}
// // NOTE: If a session is created, we also request a encryption key for now
//	//quantumElementRequest, err := c.ETSIGetQuantumInterfaces(ctx, &pb.ETSIKMSQuantumInterfaceListRequest{})
//	//if err != nil {
//	//	log.Println("Could not request quantum elements: ", err)
//	//	ph.peerStatus = KmsPeerDown
//	//	// Send notification about change
//	//	if ph.externalNotifierKMSPeer != nil {
//	//		ph.externalNotifierKMSPeer <- ph.tcpSocketStr
//	//	}
//	//	return
//	//}
//
//	//if quantumElementRequest.QlElementInfo != nil {
//	//	peerQuantumElement := quantumElementRequest.QlElementInfo[0].QleID
//	//} else {
//	//	// return error?!
//	//}
//
//	// Works and peer moves to kmsPeerUp
//	ph.peerStatus = KmsPeerUp
//
//	encryptKeyRequest, err := c.ETSIGetEncryptKeys256Bit(ctx, &pb.ETSIGetEncryptKeys256BitRequest{Amount: 1})
//	if err != nil {
//		log.Errorf("%s: could not request a encryption key: %s", kmsName, err)
//		ph.peerStatus = KmsPeerDown
//		// Send notification about change
//		if ph.externalNotifierKMSPeer != nil {
//			ph.externalNotifierKMSPeer <- ph.tcpSocketStr
//		}
//		return
//	}
//
//	// check if key is in KeyStore
//	if key, ok := ph.servingQLE.keyStorePeer.keyStore[encryptKeyRequest.KeyID]; ok {
//		keyAsString := base64.StdEncoding.EncodeToString(key.key)
//		log.Debugf("Agreed Key: %s", keyAsString)
//
//		// TODO: Remove/move ssh-kind prototype code below
//		// push the key to the encryptor via ssh
//		ssh := "ssh"
//		complexArg := fmt.Sprintf("(rc=$(sed \"12 c PresharedKey = %s\" /etc/wireguard/wg0.conf); echo \"$rc\" > /etc/wireguard/wg0.conf)", keyAsString)
//		args := []string{"root@172.20.0.4", "-oStrictHostKeyChecking=no", complexArg}
//
//		cmd := exec.Command(ssh, args...)
//		err := cmd.Run()
//		if err != nil {
//			log.Errorf("%s: could not execute ssh command with parameters", kmsName)
//			return
//		}
//	}
//
//	// Send notification about change
//	if ph.externalNotifierKMSPeer != nil {
//		ph.externalNotifierKMSPeer <- ph.tcpSocketStr
//	}
//
//	log.Printf("Greeting: %s which is now in peerStatus %d", r.GetPeerKmsName(), ph.peerStatus)
//
//	// NOTE: should be possible to cancel!
//	// By now, do check only the liveliness of the peer, nothing else.
//	//for {
//	//	select {
//	//	case <-ph.quit:
//	//		return
//	//	default:
//	//		// Contact the server and print out its response.
//	//		ctx, cancel := context.WithTimeout(context.Background(), time.Second)
//	//		defer cancel()
//	//		_, err := c.ETSICapabilities(ctx, &pb.ETSICapabilitiesRequest{MyKmsName: kmsName})
//	//		if err != nil {
//	//			log.Printf("could not greet: %v", err)
//	//			ph.peerStatus = KmsPeerDown
//	//			// Send notification about change
//	//			if ph.externalNotifierKMSPeer != nil {
//	//				ph.externalNotifierKMSPeer <- ph.tcpSocketStr
//	//			}
//	//		}
//	//		time.Sleep(30 * time.Second)
//	//	}
//	//}
//}

func (ph *kmsPeer) SyncBulkKeys(interComAddr string) error {
	rawBulkKeyIds := keysOfMap[int64](ph.servingQLE.rawBulkKeys)
	logrus.Info("FOUND BULKS: ", rawBulkKeyIds)

	ctx, cancel := context.WithTimeout(context.Background(), time.Second)
	defer cancel()
	initialPeerSetupResponse, err := ph.peerClient.SyncQkdBulk(ctx, &pbIC.SyncQkdBulkRequest{
		Timestamp:    time.Now().Unix(),
		InterComAddr: interComAddr,
		BulkId:       rawBulkKeyIds,
	})
	if err != nil {
		return err
	}

	bulkKey, ok := ph.servingQLE.rawBulkKeys[initialPeerSetupResponse.BulkId]
	if !ok {
		// TODO: add proper error message
		return fmt.Errorf("")
	}

	// TODO: Initially the peer partners should discuss about the key length,
	// for now it is hardcoded.
	ph.servingQLE.keyStorePeer = NewKmsKeyStore(256)

	if err := ph.servingQLE.keyStorePeer.KeyChopper(bulkKey); err != nil {
		return err
	}

	return nil
}

// TransportKeyNegotiation sends a request for a transport key negotiation to
// the peer KMS. The key that was agreed upon, is used to create a encrypted
// payload.
func (ph *kmsPeer) TransportKeyNegotiation() error {
	//ctx, cancel := context.WithTimeout(context.Background(), time.Second)
	//req, err := ph.peerClient.InterComTransportKeyNegotiation()
	return nil
}

func (ph *kmsPeer) SendPayload(payload []byte, pathId uuid.UUID) error {
	// select a key from key store
	key, err := ph.servingQLE.keyStorePeer.GetKey()
	if err != nil {
		return err
	}

	// Start the negotiation process for a transport key
	ctx, cancel := context.WithTimeout(context.Background(), time.Second)
	defer cancel()
	_, err = ph.peerClient.InterComTransportKeyNegotiation(ctx, &pbIC.InterComTransportKeyNegotiationRequest{
		Timestamp: time.Now().Unix(),
		PathID:    pathId.String(),
		KeyToUse:  key.keyID,
	})
	if err != nil {
		return err
	}
	// TODO: delete key?

	// TODO: would be better to update the index counter here (to keep it
	// synchronized).

	encryptedPayload, err := ph.et.Encrypt(payload, key.key)
	if err != nil {
		return err
	}

	encryptedPayloadAsString := base64.StdEncoding.EncodeToString(encryptedPayload)

	log.Infof("Sent encrypted Payload: %s", encryptedPayloadAsString)

	ctx2, cancel2 := context.WithTimeout(context.Background(), time.Second)
	defer cancel2()
	_, err = ph.peerClient.KeyForwarding(ctx2, &pbIC.KeyForwardingRequest{
		Timestamp: time.Now().Unix(),
		PathId:    pathId.String(),
		Payload:   encryptedPayloadAsString,
	})
	if err != nil {
		return err
	}

	return nil
}

func (ph *kmsPeer) GetKmsPeerStatus() KmsPeerStatus {
	return ph.peerStatus
}

func (ph *kmsPeer) GetKmsPeerId() uuid.UUID {
	return ph.id
}

func (ph *kmsPeer) GetKmsPeerQkdiId() uint32 {
	return uint32(ph.servingQLE.QlID)
}
